from __future__ import annotations

from contextlib import AsyncExitStack
from fnmatch import fnmatch
from pathlib import Path
import asyncio
import json
import os
import sys

import aiodocker
import click
import jinja2
import structlog

from bowtie import _report
from bowtie._commands import StartedDialect, TestCase
from bowtie._core import Implementation

DRAFT2020 = "https://json-schema.org/draft/2020-12/schema"
DRAFT2019 = "https://json-schema.org/draft/2019-09/schema"
DRAFT7 = "http://json-schema.org/draft-07/schema#"
DRAFT6 = "http://json-schema.org/draft-06/schema#"
DRAFT4 = "http://json-schema.org/draft-04/schema#"
DRAFT3 = "http://json-schema.org/draft-03/schema#"

DIALECT_SHORTNAMES = {
    "2020": DRAFT2020,
    "202012": DRAFT2020,
    "2020-12": DRAFT2020,
    "draft2020-12": DRAFT2020,
    "draft202012": DRAFT2020,

    "2019": DRAFT2019,
    "201909": DRAFT2019,
    "2019-09": DRAFT2019,
    "draft2019-09": DRAFT2019,
    "draft201909": DRAFT2019,

    "7": DRAFT7,
    "draft7": DRAFT7,

    "6": DRAFT6,
    "draft6": DRAFT6,

    "4": DRAFT4,
    "draft4": DRAFT4,

    "3": DRAFT3,
    "draft3": DRAFT3,
}


@click.group(context_settings=dict(help_option_names=["--help", "-h"]))
@click.version_option(prog_name="bowtie")
def main():
    """
    A meta-validator for the JSON Schema specifications.
    """

    redirect_structlog()


@main.command()
@click.argument(
    "input",
    default="-",
    type=click.File(mode="r"),
)
@click.option(
    "--out", "-o", "output",
    help="Where to write the outputted report HTML.",
    default="bowtie-report.html",
    type=click.File("w"),
)
def report(input, output):
    """
    Generate a Bowtie report from a previous run.
    """

    env = jinja2.Environment(
        loader=jinja2.PackageLoader("bowtie", "template"),
        undefined=jinja2.StrictUndefined,
        keep_trailing_newline=True,
    )
    template = env.get_template("report.html.j2")
    output.write(template.render(**_report.from_input(input)))


IMPLEMENTATION = click.option(
    "--implementation", "-i", "image_names",
    help="A docker image which implements the bowtie IO protocol.",
    multiple=True,
)
FILTER = click.option(
    "-k", "filter",
    type=lambda pattern: f"*{pattern}*",
    help="Only run cases whose description match the given glob pattern.",
)
FAIL_FAST = click.option(
    "-x", "--fail-fast",
    is_flag=True,
    default=False,
    help="Fail immediately after the first error or disagreement.",
)
SET_SCHEMA = click.option(
    "--set-schema/--no-set-schema", "-S",
    "set_schema",
    default=False,
    help=(
        "Explicitly set $schema in all (non-boolean) case schemas sent to "
        "implementations. Note this of course means what is passed to "
        "implementations will differ from what is provided in the input."
    ),
)
VALIDATE = click.option(
    "--validate-implementations/--no-validate-implementations", "-V",
    "validate_implementations",
    default=False,
    help=(
        "When speaking to implementations (provided via -i), validate "
        "(or don't validate) the requests and responses sent to them. "
        "Generally, this option protects against broken Bowtie "
        "implementations and can be left at its default (of off) unless "
        "you are developing a new implementation container."
    ),
)


@main.command()
@click.pass_context
@IMPLEMENTATION
@FILTER
@FAIL_FAST
@SET_SCHEMA
@VALIDATE
@click.option(
    "--dialect", "-D", "dialect",
    help=(
        "A URI or shortname identifying the dialect of each test case."
        f"Shortnames include: {sorted(DIALECT_SHORTNAMES)}."
    ),
    type=lambda dialect: DIALECT_SHORTNAMES.get(dialect, dialect),
    default="2020-12",
)
@click.argument(
    "input",
    default="-",
    type=click.File(mode="rb"),
)
def run(context, input, filter, **kwargs):
    """
    Run a sequence of cases provided on standard input.
    """

    cases = (TestCase.from_dict(json.loads(line)) for line in input)
    if filter:
        cases = (
            case for case in cases
            if fnmatch(case["description"], filter)
        )

    count = asyncio.run(_run(**kwargs, cases=cases))
    if not count:
        context.exit(os.EX_DATAERR)


@main.command()
@click.pass_context
@IMPLEMENTATION
@FILTER
@FAIL_FAST
@SET_SCHEMA
@VALIDATE
@click.argument(
    "input",
    default="-",
    type=click.Path(exists=True, path_type=Path),
)
def suite(context, input, filter, **kwargs):
    """
    Run a directory containing files in the official test suite format.

    Supports paths like:

        * :file:`{ROOT}/tests/draft7` to run a version's tests
        * :file:`{ROOT}/tests/draft7/foo.json` to run just one file
    """

    if input.is_dir():
        cases = suite_cases_from(files=input.glob("*.json"))
        dialect = DIALECT_SHORTNAMES.get(input.name)
    else:
        cases = suite_cases_from(files=[input])
        dialect = DIALECT_SHORTNAMES.get(input.parent.name)
    if dialect is None:
        raise click.BadParameter(
            f"{input} is not a JSON Schema Test Suite directory.",
        )

    if filter:
        cases = (
            case for case in cases
            if fnmatch(case["description"], filter)
        )

    count = asyncio.run(_run(**kwargs, dialect=dialect, cases=cases))
    if not count:
        context.exit(os.EX_DATAERR)


async def _run(
    image_names: list[str],
    cases,
    dialect: str,
    fail_fast: bool,
    set_schema: bool,
    validate_implementations: bool,
    reporter: _report.Reporter = _report.Reporter(),
):
    async with AsyncExitStack() as stack:
        docker = await stack.enter_async_context(aiodocker.Docker())

        starting = [
            stack.enter_async_context(
                Implementation.start(
                    docker=docker,
                    image_name=image_name,
                    validate_implementations=validate_implementations,
                ),
            ) for image_name in image_names
        ]
        reporter.will_speak(dialect=dialect)
        implementations = []
        for each in asyncio.as_completed(starting):
            implementation = await each

            if implementation.supports_dialect(dialect):
                response = await implementation.start_speaking(dialect)
                if response != StartedDialect.OK:
                    reporter.unacknowledged_dialect(
                        implementation=implementation,
                        dialect=dialect,
                        response=response,
                    )
                implementations.append(implementation)
            else:
                reporter.unsupported_dialect(
                    implementation=implementation,
                    dialect=dialect,
                )
        reporter.ready(implementations=implementations, dialect=dialect)

        seq = 0
        should_stop = False
        for seq, case, case_reporter in sequenced(cases, reporter):
            if set_schema and not isinstance(case["schema"], bool):
                case["schema"]["$schema"] = dialect

            responses = [
                each.run_case(seq=seq, case=case) for each in implementations
            ]
            for each in asyncio.as_completed(responses):
                response = await each
                response.report(reporter=case_reporter)
                if fail_fast and not response.succeeded:
                    # Stop after this case, since we still have awaitables out
                    should_stop = True

            if should_stop:
                break
        reporter.finished(count=seq)
    return seq


def sequenced(cases, reporter):
    for seq, case in enumerate(cases, 1):
        yield seq, case, reporter.case_started(seq=seq, case=case)


def suite_cases_from(files):
    for file in files:
        for case in json.loads(file.read_text()):
            for test in case["tests"]:
                test["instance"] = test.pop("data")
            yield TestCase.from_dict(case)


def redirect_structlog(file=sys.stderr):
    """
    Reconfigure structlog's defaults to go to the given location.
    """

    structlog.configure(
        processors=[
            structlog.processors.add_log_level,
            structlog.processors.StackInfoRenderer(),
            structlog.dev.set_exc_info,
            structlog.processors.TimeStamper(
                fmt="%Y-%m-%d %H:%M.%S", utc=False,
            ),
            structlog.dev.ConsoleRenderer(
                colors=getattr(file, "isatty", lambda: False)(),
            ),
        ],
        logger_factory=structlog.PrintLoggerFactory(file),
    )



"""
Hand crafted classes which should undoubtedly be autogenerated from the schema.
"""
from __future__ import annotations

import attrs

from bowtie import exceptions


@attrs.define
class Test:

    description: str
    instance: object
    valid: bool | None = None


@attrs.define
class TestCase:

    description: str
    schema: object
    tests: list[Test]
    comment: str | None = None

    @classmethod
    def from_dict(cls, data):
        data["tests"] = [Test(**test) for test in data["tests"]]
        return cls(**data)

    def without_expected_results(self):
        as_dict = {
            "tests": [
                attrs.asdict(test, filter=lambda k, _: k.name != "valid")
                for test in self.tests
            ],
        }
        as_dict.update(
            attrs.asdict(
                self,
                filter=lambda k, v: k.name != "tests" and (
                    k.name != "comment" or v is not None
                ),
            ),
        )
        return as_dict


@attrs.define
class Started:

    succeeded = True

    implementation: dict
    ready: bool = False
    version: int = None


@attrs.define
class Start:

    cmd = "start"
    Response = Started

    version: int

    def succeed(self, response):
        if not response.succeeded:
            raise exceptions.StartupFailure(self, response.stderr)
        elif not response.ready:
            raise exceptions.ImplementationNotReady(self)
        else:
            version = response.version
            if version != 1:
                raise exceptions.VersionMismatch(self, expected=1, got=version)

        return response.implementation


START_V1 = Start(version=1)


@attrs.define
class StartedDialect:

    succeeded = True

    ok: bool


StartedDialect.OK = StartedDialect(ok=True)


@attrs.define
class Dialect:

    cmd = "dialect"
    Response = StartedDialect

    dialect: str

    def succeed(self, response):
        return response


def _case_result(errored=False, **response):
    if errored:
        return lambda implementation, expected: CaseErrored(
            implementation=implementation,
            **response,
        )
    return lambda implementation, expected: CaseResult(
        implementation=implementation,
        expected=expected,
        **response,
    )


@attrs.define
class CaseResult:
    """
    The result of running a test case.
    """

    succeeded = True

    implementation: str
    seq: int
    results: dict
    expected: list[bool | None]

    def report(self, reporter):
        reporter.got_results(self)


@attrs.define
class CaseErrored:

    succeeded = False

    implementation: str
    seq: int
    context: dict

    def report(self, reporter):
        return reporter.errored(self.implementation, self.context)


@attrs.define
class Run:

    cmd = "run"
    Response = staticmethod(_case_result)

    seq: int
    case: dict

    def succeed(self, response):
        return response


@attrs.define
class Stop:

    cmd = "stop"

    def Response(self, **kwargs):
        # FIXME: Ugh, the last response... we're off by one.
        if kwargs:
            return _case_result(**kwargs)

    def succeed(self, response):
        return response


STOP = Stop()


def to_request(command):
    as_dict = {"cmd": command.cmd}
    as_dict.update(attrs.asdict(command))
    return as_dict



from __future__ import annotations

from contextlib import asynccontextmanager
from importlib import resources
import asyncio
import json

import aiodocker
import attrs

from bowtie import _commands


@attrs.define
class InvalidResponse:
    """
    An implementation sent an invalid response to a command.
    """

    succeeded = False

    exc_info: Exception
    response: dict


@attrs.define
class BackingOff:
    """
    An implementation has failed too many times.
    """

    succeeded = False

    implementation: str

    def report(self, reporter):
        reporter.backoff(implementation=self.implementation)


@attrs.define
class UncaughtError:
    """
    An implementation spewed to its stderr.
    """

    implementation: str
    stderr: bytes

    succeeded = False

    def report(self, reporter):
        reporter.errored_uncaught(
            implementation=self.implementation,
            stderr=self.stderr,
        )


@attrs.define
class BadFraming:
    """
    We're confused about line endings.
    """

    data: bytes

    succeeded = False

    def report(self, reporter):
        reporter.errored_uncaught(
            implementation=self.implementation,
            data=self.data,
        )


@attrs.define
class Empty:
    """
    We didn't get a response.
    """

    implementation: str

    succeeded = False

    def report(self, reporter):
        reporter.errored_uncaught(
            implementation=self.implementation,
            reason="Empty response.",
        )


@attrs.define(hash=True, slots=False)
class Implementation:
    """
    A running implementation under test.
    """

    name: str

    _maybe_validate: callable

    _docker: aiodocker.Docker = attrs.field(repr=False)
    _restarts: int = attrs.field(default=20 + 1, repr=False)
    _read_timeout_sec: float = attrs.field(default=2.0, repr=False)

    _container: aiodocker.containers.DockerContainer = attrs.field(
        default=None, repr=False,
    )
    _stream: aiodocker.stream.Stream = attrs.field(default=None, repr=False)

    metadata: dict = {}

    _dialect = None

    @classmethod
    @asynccontextmanager
    async def start(cls, docker, image_name, validate_implementations):
        if validate_implementations:
            from jsonschema.validators import RefResolver, validator_for

            text = resources.read_text("bowtie.schemas", "io-schema.json")
            root_schema = json.loads(text)
            resolver = RefResolver.from_schema(root_schema)
            Validator = validator_for(root_schema)
            Validator.check_schema(root_schema)

            def validate(instance, schema):
                resolver.store["urn:current-dialect"] = {"$ref": self._dialect}
                Validator(schema, resolver=resolver).validate(instance)
        else:
            def validate(instance, schema):
                pass

        self = cls(name=image_name, docker=docker, maybe_validate=validate)
        try:
            await self._restart_container()
            yield self
            await self._stop()
        finally:
            try:
                await self._container.delete(force=True)
            except aiodocker.exceptions.DockerError:
                pass

    async def _restart_container(self):
        self._restarts -= 1

        if self._container is not None:
            await self._container.delete(force=True)
        self._container = await self._docker.containers.create(
            config=dict(Image=self.name, OpenStdin=True),
        )
        await self._container.start()
        self._stream = self._container.attach(
            stdin=True,
            stdout=True,
            stderr=True,
        )
        self.metadata = await self._send(_commands.START_V1)

    def supports_dialect(self, dialect):
        return dialect in self.metadata.get("dialects", [])

    async def start_speaking(self, dialect):
        self._dialect = dialect
        return await self._send(_commands.Dialect(dialect=dialect))

    async def run_case(self, seq, case):
        if self._restarts <= 0:
            return BackingOff(implementation=self.name)

        command = _commands.Run(seq=seq, case=case.without_expected_results())
        response = await self._send(command)
        expected = [test.valid for test in case.tests]
        return response(implementation=self.name, expected=expected)

    async def _stop(self):
        if self._restarts > 0:
            await self._send(_commands.STOP)

    async def _send(self, cmd, retry=3):
        request = _commands.to_request(cmd)
        pointer = f"#/$defs/command/$defs/{cmd.cmd}"
        self._maybe_validate(instance=request, schema={"$ref": pointer})
        as_bytes = f"{json.dumps(request)}\n".encode("utf-8")

        try:
            await self._stream.write_in(as_bytes)
        except AttributeError:
            # FIXME: aiodocker doesn't appear to properly report when its
            # stream is closed
            await self._restart_container()
            await self._stream.write_in(as_bytes)

        for _ in range(retry):
            try:
                message = await self._read_with_timeout()
                if message is None:
                    continue
            except asyncio.exceptions.TimeoutError:
                continue

            if message.stream == 2:
                data = []
                while message is not None and message.stream == 2:
                    data.append(message.data)
                    try:
                        message = await self._read_with_timeout()
                    except asyncio.exceptions.TimeoutError:
                        break
                return lambda *args, **kwargs: UncaughtError(
                    implementation=self.name,
                    stderr=b"".join(data),  # noqa: B023
                )

            data = message.data
            assert b"\n" not in message[:-1]
            while not data.endswith(b"\n"):
                try:
                    message = await self._read_with_timeout()
                except asyncio.exceptions.TimeoutError:
                    return lambda *args, **kwargs: BadFraming(  # noqa: B023
                        implementation=self.name,
                        data=b"".join(data),  # noqa: B023
                    )
                data += message.data
            try:
                data = json.loads(data)
            except json.JSONDecodeError:
                return lambda *args, **kwargs: BadFraming(
                    implementation=self.name,
                    data=b"".join(data),  # noqa: B023
                )
            else:
                pointer = f"#/$defs/command/$defs/{cmd.cmd}/$defs/response"
                self._maybe_validate(instance=data, schema={"$ref": pointer})
                return cmd.succeed(cmd.Response(**data))
        return lambda *args, **kwargs: Empty(implementation=self.name)

    def _read_with_timeout(self):
        return asyncio.wait_for(
            self._stream.read_out(),
            timeout=self._read_timeout_sec,
        )



from pathlib import Path

import nox

ROOT = Path(__file__).parent
DOCS = ROOT / "docs"
BOWTIE = ROOT / "bowtie"
TESTS = ROOT / "tests"


@nox.session(python=["3.8", "3.9", "3.10"])
def tests(session):
    session.install("-r", str(ROOT / "test-requirements.txt"))
    session.run("pytest")


@nox.session(tags=["build"])
def build(session):
    session.install("build")
    tmpdir = session.create_tmp()
    session.run("python", "-m", "build", str(ROOT), "--outdir", tmpdir)


@nox.session(tags=["build"])
def shiv(session):
    session.install("shiv")
    if session.posargs:
        out = session.posargs[0]
    else:
        tmpdir = Path(session.create_tmp())
        out = tmpdir / "bowtie"
    session.run(
        "python", "-m", "shiv",
        "--reproducible",
        "-c", "bowtie",
        str(ROOT),
        "-o", str(out),
    )
    print(f"Outputted a shiv to {out}.")


@nox.session(tags=["style"])
def readme(session):
    session.install("build", "twine")
    tmpdir = session.create_tmp()
    session.run("python", "-m", "build", str(ROOT), "--outdir", tmpdir)
    session.run("python", "-m", "twine", "check", tmpdir + "/*")


@nox.session(tags=["style"])
def style(session):
    session.install(
        "flake8",
        "flake8-broken-line",
        "flake8-bugbear",
        "flake8-commas",
        "flake8-quotes",
        "flake8-tidy-imports",
    )
    session.run("python", "-m", "flake8", str(BOWTIE), str(TESTS), __file__)


@nox.session(tags=["docs"])
@nox.parametrize(
    "builder", [
        nox.param(name, id=name) for name in [
            "dirhtml", "doctest", "linkcheck", "spelling",
        ]
    ],
)
def docs(session, builder):
    session.install("-r", str(DOCS / "requirements.txt"))
    tmpdir = Path(session.create_tmp())
    argv = ["-n", "-T", "-W"]
    if builder != "spelling":
        argv += ["-q"]
    session.run(
        "python", "-m", "sphinx", "-b", builder,
        str(DOCS), str(tmpdir / builder), *argv,
    )


@nox.session(tags=["docs", "style"], name="docs(style)")
def docs_style(session):
    session.install(
        "doc8",
        "pygments",
        "pygments-github-lexers",
    )
    session.run("python", "-m", "doc8", "--max-line-length", "1000", str(DOCS))



import json
import sys

import attrs
import structlog

from bowtie._commands import TestCase


def writer(file=sys.stdout):
    return lambda **result: file.write(f"{json.dumps(result)}\n")


@attrs.define
class Reporter:

    _write = attrs.field(default=writer())
    _log: structlog.BoundLogger = attrs.field(factory=structlog.get_logger)

    def unsupported_dialect(self, implementation, dialect):
        self._log.warn(
            "Unsupported dialect, skipping implementation.",
            logger_name=implementation.name,
            dialect=dialect,
        )

    def unacknowledged_dialect(self, implementation, dialect, response):
        self._log.warn(
            (
                "Implicit dialect not acknowledged. "
                "Proceeding, but implementation may not have configured "
                "itself to handle schemas without $schema."
            ),
            logger_name=implementation.name,
            dialect=dialect,
            response=response,
        )

    def ready(self, implementations, dialect):
        metadata = {
            implementation.name: dict(
                implementation.metadata, image=implementation.name,
            ) for implementation in implementations
        }
        self._write(implementations=metadata)

    def will_speak(self, dialect):
        self._log.info("Will speak dialect", dialect=dialect)

    def finished(self, count):
        if not count:
            self._log.error("No test cases ran.")
        else:
            self._log.msg("Finished", count=count)

    def case_started(self, seq, case):
        return _CaseReporter.case_started(
            case=case,
            seq=seq,
            write=self._write,
            log=self._log.bind(
                seq=seq,
                case=case.description,
                schema=case.schema,
            ),
        )


@attrs.define
class _CaseReporter:

    _write: callable
    _log: structlog.BoundLogger

    @classmethod
    def case_started(cls, log, write, case: TestCase, seq: int):
        self = cls(log=log, write=write)
        self._write(case=attrs.asdict(case), seq=seq)
        return self

    def got_results(self, results):
        self._write(**attrs.asdict(results))

    def backoff(self, implementation):
        self._log.warn("backing off", logger_name=implementation)

    def errored(self, implementation, response):
        self._log.error("", logger_name=implementation, **response)

    def errored_uncaught(self, implementation, **response):
        self._log.error("uncaught", logger_name=implementation, **response)


def from_input(input):
    """
    Create a structure suitable for the report template from an input file.
    """

    lines = (json.loads(line) for line in input)
    header = next(lines)
    implementations = header["implementations"]

    combined = {}

    for each in lines:
        if "case" in each:
            combined[each["seq"]] = {
                "case": each["case"],
                "results": [(test, {}) for test in each["case"]["tests"]],
            }
            continue

        implementation = each.pop("implementation")
        case = combined[each["seq"]]

        for result, expected, (_, seen) in zip(
            each["results"],
            each["expected"],
            case["results"],
        ):
            incorrect = expected is not None and result["valid"] != expected
            seen[implementation] = result, incorrect

    return dict(
        implementations=implementations.values(),
        results=[(v, k) for k, v in sorted(combined.items())],
    )
